""" Command Line Script Support.

    Copyright (c) 2009, 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import errno
import logging.config
import os
import signal
import sys
import textwrap
import time
import traceback

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from typing import List, Optional

import shtab

from pyrosimple import error
from pyrosimple.util import pymagic


class ScriptBase:
    """Base class for command line interfaces."""

    # additonal stuff appended after the command handler's docstring
    ADDITIONAL_HELP: List[str] = []

    def __init__(self):
        """Initialize CLI."""
        self.startup = time.time()
        # self.LOG exists only for backwards compatibility
        self.LOG = self.log = pymagic.get_class_logger(self)

        self.args = None
        self.options = None
        self.return_code = 0
        self.engine = None
        self.intermixed_args = False

        logging.basicConfig(level=logging.WARNING)
        # For python 3.7 compatibility
        try:
            import importlib.metadata  # pylint: disable=import-outside-toplevel

            version = importlib.metadata.version(  # pylint: disable=no-member
                "pyrosimple"
            )
        except ImportError:
            version = "unknown"
        implementation = sys.implementation.name
        if implementation == "cpython":
            implementation = "Python"
        version_info = f"{version} on {implementation} {sys.version.split()[0]}"

        self.parser = ArgumentParser(
            formatter_class=RawDescriptionHelpFormatter,
            description="%(prog)s "
            + version_info
            + "\n\n"
            + textwrap.dedent(self.__doc__.rstrip()).lstrip("\n")
            + "\n".join(self.ADDITIONAL_HELP)
            + "\n\nFor more details, see the full documentation at"
            + "\n    https://kannibalox.github.io/pyrosimple/",
        )
        shtab.add_argument_to(self.parser, ["--print-completion"])

        self.parser.add_argument(
            "--version", action="version", version=f"%(prog)s {version_info}"
        )

    def add_bool_option(self, *args, **kwargs):
        """Add a boolean option.

        @keyword help: Option description.
        """
        dest = (
            [o for o in args if o.startswith("--")][0]
            .replace("--", "")
            .replace("-", "_")
        )
        self.parser.add_argument(
            dest=dest, action="store_true", default=False, help=kwargs["help"], *args
        )

    def add_value_option(self, *args, **kwargs):
        """Add a value option.

        @keyword dest: Destination attribute, derived from long option name if not given.
        @keyword action: How to handle the option.
        @keyword help: Option description.
        @keyword default: If given, add this value to the help string.
        """
        kwargs["metavar"] = args[-1]
        if "dest" not in kwargs:
            kwargs["dest"] = (
                [o for o in args if o.startswith("--")][0]
                .replace("--", "")
                .replace("-", "_")
            )
        if "default" in kwargs and kwargs["default"]:
            kwargs["help"] += f" [{kwargs['default']}]"
        if "choices" in kwargs:
            del kwargs["type"]
        self.parser.add_argument(*args[:-1], **kwargs)

    def get_options(self):
        """Get program options."""
        self.parser.add_argument(
            "-q",
            "--quiet",
            "--cron",  # For backwards compatibility
            help="silence warnings",
            dest="log_level",
            action="store_const",
            default=logging.WARNING,
            const=logging.CRITICAL,
        )
        self.parser.add_argument(
            "--debug",
            help="show detailed messages",
            dest="log_level",
            action="store_const",
            const=logging.DEBUG,
        )
        self.parser.add_argument(
            "-v",
            "--verbose",
            help="show additional information",
            dest="log_level",
            action="store_const",
            const=logging.INFO,
        )

        # Template method to add options of derived class
        self.add_options()

        if self.intermixed_args:
            self.options = self.parser.parse_intermixed_args(self.args)
        else:
            self.options = self.parser.parse_args(self.args)
        self.args: Optional[List] = getattr(self.options, "args", None)

        if self.options.log_level:
            logging.getLogger("pyrosimple").setLevel(self.options.log_level)

        self.log.debug(
            "Options: %s",
            ", ".join("%s=%r" % i for i in sorted(vars(self.options).items())),
        )

    def fatal(self, msg, exc=None):
        """Exit on a fatal error."""
        if exc is not None:
            self.log.fatal("%s (%s)", msg, exc)
            if logging.getLogger().isEnabledFor(logging.DEBUG):
                return  # let the caller re-raise it
        else:
            self.log.fatal(msg)
        sys.exit(error.EX_SOFTWARE)

    def run(self, args=None):
        """The main program skeleton."""
        self.args = args
        try:
            try:
                # Preparation steps
                self.get_options()

                # Template method with the tool's main loop
                self.mainloop()
            except error.LoggableError:
                print(traceback.format_exc())
                sys.exit(error.EX_SOFTWARE)
            except KeyboardInterrupt:
                print()
                self.log.critical("Aborted by CTRL-C!\n")
                signal.signal(signal.SIGINT, signal.SIG_DFL)
                os.kill(os.getpid(), signal.SIGINT)
            except OSError as exc:
                # [Errno 32] Broken pipe?
                if exc.errno == errno.EPIPE:
                    print(f"\n{exc}, exiting!\n", file=sys.stderr)
                    sys.exit(error.EX_IOERR)
                else:
                    raise
        finally:
            # Shut down
            if self.options:  ## No time logging on --version and such
                running_time = time.time() - self.startup
                self.log.info("Total time: %.3f seconds.", running_time)
            logging.shutdown()

        # Special exit code?
        if self.return_code:
            sys.exit(self.return_code)

    def add_options(self):
        """Add program options."""

    def mainloop(self):
        """The main loop."""
        raise NotImplementedError()

    def rpc_stats(self) -> str:
        """Return a string with RPC statistics"""

        from pyrosimple import io  # pylint: disable=import-outside-toplevel
        from pyrosimple.util import fmt  # pylint: disable=import-outside-toplevel

        req_num = int(io.scgi.request_counter._value.get())
        req_sz = fmt.human_size(io.scgi.request_size_counter._value.get())
        req_time = round(
            list(io.scgi.response_time_summary._child_samples())[1].value, 3
        )
        resp_sz = fmt.human_size(io.scgi.response_size_counter._value.get())
        return f"{req_num} requests ({req_sz}) in {req_time}s (response {resp_sz})"


class ScriptBaseWithConfig(ScriptBase):  # pylint: disable=abstract-method
    """CLI tool with configuration support."""

    def add_options(self):
        super().add_options()
        self.parser.add_argument("-U", "--url", help="URL to rtorrent instance")

    def get_options(self):
        """Get program options."""
        super().get_options()
        # pylint: disable=import-outside-toplevel
        from pyrosimple import config
        from pyrosimple.torrent import rtorrent

        # pylint: enable=import-outside-toplevel

        if self.options.url:
            config.settings["SCGI_URL"] = config.lookup_connection_alias(
                self.options.url
            )
        config.load_custom_py()
        self.engine = rtorrent.RtorrentEngine()



""" Logging Support.

    Copyright (c) 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import logging


def get_logfile(logger=None):
    """Return log file of first file handler associated with the (root) logger.
    None if no such handler is found.
    """
    logger = logger or logging.getLogger()
    handlers = [i for i in logger.handlers if isinstance(i, logging.FileHandler)]
    return handlers[0].baseFilename if handlers else None



""" rTorrent Proxy.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import base64
import errno
import logging
import operator
import os
import shlex
import time
import urllib.parse

from functools import lru_cache, partial
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional, Set, Tuple, Union
from xmlrpc import client as xmlrpclib

import bencode
import jinja2

from pyrosimple import config, error
from pyrosimple.torrent import engine
from pyrosimple.util import fmt, matching, metafile, pymagic, rpc, traits
from pyrosimple.util.cache import ExpiringCache
from pyrosimple.util.parts import Bunch


# Prepare the jinja template environment at the module level
env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        [Path("~/.config/pyrosimple/templates/").expanduser()]
    ),
)
# Load filter methods from fmt submodule
env.filters.update(
    {
        name[4:]: method
        for name, method in fmt.__dict__.items()
        if name.startswith("fmt_")
    }
)


class CommaLexer(shlex.shlex):
    """Helper to split argument lists."""

    def __init__(self, text: str):
        shlex.shlex.__init__(self, text, None, True)
        self.whitespace += ","
        self.whitespace_split = True
        self.commenters = ""


class RtorrentItem(engine.TorrentProxy):
    """A single download item."""

    def __init__(
        self,
        engine_,
        fields,
        rpc_fields: Optional[Dict] = None,
        cache_expires: Optional[float] = None,
    ):
        """Initialize download item."""
        super().__init__()
        if cache_expires is None:
            cache_expires = float(config.settings.ITEM_CACHE_EXPIRATION)
        self._engine = engine_
        self._fields = ExpiringCache(
            static_keys=engine.FieldDefinition.CONSTANT_FIELDS, expires=cache_expires
        )  # Acts a cache for the item
        self._fields.update(dict(fields))
        self._rpc_cache = ExpiringCache(
            static_keys={
                "d.name",
                "d.size_bytes",
                "d.size_chunks",
            }
        )
        if rpc_fields is not None:
            self._rpc_cache.update(rpc_fields)
        if "hash" not in fields:
            self._fields["hash"] = self.rpc_call("d.hash")

    def _make_it_so(
        self, command: str, calls: List[str], *args, observer: Optional[Callable] = None
    ):
        """Perform some error-checked RPC calls."""
        args = (self._fields["hash"],) + args
        try:
            self._engine.logger.debug(
                "%s%s torrent %s",
                command[0].upper(),
                command[1:],
                self._fields["hash"],
            )
            for call in calls:
                namespace = self._engine.rpc
                result = getattr(namespace, call.lstrip(":"))(*args)
                if observer is not None:
                    observer(result)
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While {command} torrent {self._fields['hash']}: {exc}"
            ) from exc

    def _get_files(self, attrs: Optional[List[str]] = None):
        """Get a list of all files in this download; each entry has the
        attributes C{path} (relative to root), C{size} (in bytes),
        C{mtime}, C{prio} (0=off, 1=normal, 2=high), C{created},
        and C{opened}.

        @param attrs: Optional list of additional attributes to fetch.
        """
        try:
            # Get info for all files
            f_multicall = self._engine.rpc.f.multicall
            f_params = [
                self._fields["hash"],
                rpc.NOHASH,
                "f.path=",
                "f.size_bytes=",
                "f.last_touched=",
                "f.priority=",
                "f.is_created=",
                "f.is_open=",
            ]
            for attr in attrs or []:
                f_params.append(f"f.{attr}=")
            rpc_result = f_multicall(*tuple(f_params))
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting files for torrent #{self._fields['hash']}: {exc}"
            )
        else:
            # self._engine.logger.debug("files result: %r" % rpc_result)

            # Return results
            result = [
                Bunch(
                    path=i[0],
                    size=i[1],
                    mtime=i[2] / 1000000.0,
                    prio=i[3],
                    created=i[4],
                    opened=i[5],
                )
                for i in rpc_result
            ]

            if attrs:
                for idx, attr in enumerate(attrs):
                    for item, rpc_item in zip(result, rpc_result):
                        item[attr] = rpc_item[6 + idx]

            return result

    def memoize(self, name: str, getter: Callable, *args, **kwargs):
        """Cache a stable expensive-to-get item value for later
        (optimized) retrieval."""
        field = "custom_memo_" + name
        cached = self.rpc_call("d.custom", ["memo_" + name])
        if cached:
            value = cached
        else:
            value = getter(*args, **kwargs)
            self._make_it_so(
                f"caching {name}={value!r} for",
                ["d.custom.set"],
                field[7:],
                value,
            )
            self._fields[field] = value
        return value

    def _get_kind(self, limit) -> Set[str]:
        """Get a set of dominant file types. The files must contribute
        at least C{limit}% to the item's total size.
        """
        histo = self.rpc_call("d.custom", ["kind"])

        if histo:
            # Parse histogram from cached field
            histo = [i.split("%_") for i in str(histo).split()]
            histo = [(int(val, 10), ext) for val, ext in histo]
        else:
            # Get file types
            histo = traits.get_filetypes(
                self._get_files(),
                path=operator.attrgetter("path"),
                size=operator.attrgetter("size"),
            )

            # Set custom cache field with value formatted like "80%_flac 20%_jpg" (sorted by percentage)
            histo_str = " ".join([f'{i[0]}%_{i[1].replace(" ", "_")}' for i in histo])
            self._make_it_so(
                f"setting kind cache {histo_str!r} on",
                ["d.custom.set"],
                "kind",
                histo_str,
            )
            self._fields["custom_kind"] = histo_str

        # Return all non-empty extensions that make up at least <limit>% of total size
        return {ext for val, ext in histo if ext and val >= limit}

    def as_dict(self):
        """Return known fields."""
        return dict(self._fields)

    def rpc_call(self, method: str, args: Optional[List] = None, cache: bool = True):
        """Directly call rpc for item-specific information"""
        cache_key = method
        if args:
            cache_key += "=" + ",".join([str(a) for a in args])
        val = self._rpc_cache.get(cache_key, None)
        if cache and val is not None:
            return val
        if args is None:
            args = []
        getter = getattr(self._engine.rpc, method)
        val = getter(self._fields["hash"], *args)
        self._rpc_cache[cache_key] = val
        return val

    def fetch(self, name: str, cache: bool = True):
        """Get a field on demand. By 'on demand', this means that the field may possibly be created
        if it does not already exists (e.g. custom fields). It also allows directly controlling if the _fields cache
        should be used"""
        if cache:
            try:
                return self._fields[name]
            except KeyError:
                pass
        if engine.TorrentProxy.add_manifold_attribute(name) is None:
            raise AttributeError(name)
        value = getattr(self, name)
        self._fields[name] = value
        return value

    def __getattr__(self, name):
        return self.fetch(name)

    def datapath(self) -> Path:
        """Get an item's data path."""
        directory = self.rpc_call("d.directory")
        if self.rpc_call("d.is_multi_file"):
            path = Path(directory)
        else:
            path = Path(directory, self.rpc_call("d.name"))
        return path.expanduser()

    def announce_urls(self, default=[]):  # pylint: disable=dangerous-default-value
        """Get a list of all announce URLs.
        Returns `default` if no trackers are found at all.
        """
        try:
            response = self.rpc_call("t.multicall", ["", "t.url=", "t.is_enabled="])
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting announce URLs for #{self._fields['hash']}: {exc}"
            )

        if response:
            return [i[0] for i in response if i[1]]
        return default

    def start(self):
        """(Re-)start downloading or seeding."""
        self._make_it_so("starting", ["d.open", "d.start"])

    def stop(self):
        """Stop and close download."""
        self._make_it_so("stopping", ["d.stop", "d.close"])

    def ignore(self, flag: int):
        """Set ignore status."""
        self._make_it_so("setting ignore status for", ["d.ignore_commands.set"], flag)
        self._fields["is_ignored"] = flag

    def set_prio(self, prio: int):
        """Set priority (0-3)."""
        self._make_it_so(
            "setting priority for", ["d.priority.set"], str(max(0, min(int(prio), 3)))
        )

    def tag(self, tags: str):
        """Add or remove tags."""
        # Get tag list and add/remove given tags
        tags = tags.lower()

        previous: List[str] = self.rpc_call("d.custom", ["tags"]).split()
        tagset = set(previous)
        for tag in tags.replace(",", " ").split():
            if tag.startswith("-"):
                tagset.discard(tag[1:])
            elif tag.startswith("+"):
                tagset.add(tag[1:])
            elif tag:
                tagset.add(tag)

        # Write back new tagset, if changed
        tagset.discard("")
        if list(tagset) != previous:
            new_tags = " ".join(sorted(tagset))
            self._make_it_so(
                f"setting changed tags {tagset!r} on",
                ["d.custom.set"],
                "tags",
                new_tags,
            )
            self._fields["custom_tags"] = new_tags

    def set_throttle(self, name: str):
        """Assign to throttle group."""
        if name.lower() == "null":
            name = "NULL"
        if name.lower() == "none":
            name = ""

        if name not in self._engine.known_throttle_names:
            if self._engine.rpc.throttle.up.max(rpc.NOHASH, name) == -1:
                if self._engine.rpc.throttle.down.max(rpc.NOHASH, name) == -1:
                    raise error.UserError(f"Unknown throttle name '{name}'")
            self._engine.known_throttle_names.add(name)

        if (name or "NONE") == self.rpc_call("d.throttle_name"):
            self._engine.logger.debug(
                "Keeping throttle %r on torrent #%s",
                self.rpc_call("d.throttle_name"),
                self._fields["hash"],
            )
            return

        active = self.rpc_call("d.is_active")
        if active:
            self._engine.logger.debug(
                "Torrent #%s stopped for throttling", self._fields["hash"]
            )
            self.stop()
        self._make_it_so(f"setting throttle {name!r} on", ["d.throttle_name.set"], name)
        if active:
            self._engine.logger.debug(
                "Torrent #%s restarted after throttling",
                self._fields["hash"],
            )
            self.start()

    def set_custom(self, key: str, value: Optional[str] = None):
        """Set a custom value. C{key} might have the form "key=value" when value is C{None}."""
        # Split combined key/value
        if value is None:
            try:
                key, value = key.split("=", 1)
            except (ValueError, TypeError) as exc:
                raise error.UserError(
                    f"Bad custom field assignment {key!r}, probably missing a '=' ({exc})"
                )

        # Check identifier rules
        args: List[str]
        if not key:
            raise error.UserError("Custom field name cannot be empty!")
        if len(key) == 1 and key in "12345":
            method, args = "d.custom" + key + ".set", [value]
        elif not (key[0].isalpha() and key.replace("_", "").isalnum()):
            raise error.UserError(
                f"Bad custom field name {key!r} (must only contain a-z, A-Z, 0-9 and _)"
            )
        else:
            method, args = "d.custom.set", [key, value]

        # Make the assignment
        self._make_it_so(f"setting custom_{key} = {value!r} on", [method], *args)
        self._fields["custom_" + key] = value

    def hash_check(self):
        """Hash check a download."""
        self._make_it_so("hash-checking", ["d.check_hash"])

    def __print_result(self, data, method=None, args=None):
        "Helper to print RPC call results"
        args_list = ""
        if args:
            args_list = '"' + '","'.join(args) + '"'
        print(f"{self._fields['hash']}\t{data}\t{method.lstrip(':')}={args_list}")

    def execute(self, commands):
        """Execute RPC command(s)."""
        try:
            commands = [i.strip() for i in commands.split(" ; ")]
        except (TypeError, AttributeError):
            pass  # assume an iterable

        for command in commands:
            try:
                method, args = command.split("=", 1)
                args = tuple(CommaLexer(args))
            except (ValueError, TypeError) as exc:
                raise error.UserError(
                    f"Bad command {command!r}, probably missing a '=' ({exc})"
                )

            observer = None
            if method.startswith(">"):
                observer = partial(self.__print_result, args=args, method=method)
            method = method.lstrip(">")
            self._make_it_so("executing command on", [method], *args, observer=observer)

    def custom_items(self) -> Dict:
        """Try using rtorrent-ps commands to list custom keys, otherwise fall back to reading from a session file.

        This does *not* include the custom1, custom2, etc. keys"""
        proxy = self._engine.open()
        if self._engine.has_method("d.custom.keys"):
            custom_fields = {}
            for key in proxy.d.custom.keys(self.hash):
                custom_fields[key] = proxy.d.custom(self.hash, key)
            return custom_fields
        proxy.d.save_full_session(self.hash)
        info_file = Path(proxy.session.path(), f"{self.hash}.torrent.rtorrent")
        return dict(
            bencode.decode(proxy.execute.capture(rpc.NOHASH, "cat", info_file))[
                "custom"
            ]
        )

    def move(self, dest: os.PathLike, move_func: Optional[Callable] = None):
        """Move files from one path to another. By default it will do
        a simple move of only related files while replicating the same
        directory structure, but `move_func` allows providing custom
        behavior"""
        if move_func is None:

            def _default_move(_item, src, dest):
                import shutil  # pylint: disable=import-outside-toplevel

                shutil.move(src, dest)

            move_func = _default_move
        if self.rpc_call("d.is_multi_file"):
            for f in self._get_files():
                src = Path(self.datapath(), f.path)
                move_func(self, src, Path(dest, f.path))
        else:
            move_func(self, self.datapath(), dest)

    def move_to_host(self, remote_url: str, copy: bool = False):
        """Migrate an item to a remote host"""

        # TODO Generalize this overriding of query parameters
        parsed_url = urllib.parse.urlsplit(remote_url)
        queries = urllib.parse.parse_qs(parsed_url.query)
        rpc_protocol = queries.get("rpc", ["xml"])[0]
        remote_proxy = RtorrentEngine(remote_url).open()
        proxy = self._engine.open()
        self._engine.logger.debug("Attempting to move %s to %s", self.hash, remote_url)
        extra_cmds: List[str] = []
        # Check if hash already exists remotely
        try:
            remote_proxy.d.hash(self.hash)
        except rpc.HashNotFound:
            pass
        else:
            raise error.EngineError(
                f"Hash {self.hash} already exists remotely on {remote_url}"
            )
        # This might be brittle for systems that have a low
        # network.xmlrpc.size_limit but large torrents.
        torrent_path = Path(proxy.session.path(), f"{self.hash}.torrent")
        torrent = metafile.Metafile(
            bencode.decode(
                base64.b64decode(
                    proxy.execute.capture(
                        rpc.NOHASH,
                        "base64",
                        str(torrent_path),
                    )
                )
            )
        )
        try:
            torrent.add_fast_resume(Path(proxy.d.directory_base(self.hash)))
        except (FileNotFoundError, OSError) as e:
            self._engine.logger.error("Could not add fast resume data: %s", e)
        # Do some basic escaping, nothing else should be necessary.
        base_dir = proxy.d.directory_base(self.hash).replace('"', r"\"")
        extra_cmds.insert(0, f'd.directory_base.set="{base_dir}"')
        rpc_metafile = xmlrpclib.Binary(bencode.bencode(dict(torrent)))
        if not copy:
            proxy.d.stop(self.hash)
        self._engine.logger.debug("Running extra commands on load: %s", extra_cmds)
        if rpc_protocol == "json":
            remote_proxy.load.verbose("", rpc_metafile, *extra_cmds)
        else:
            remote_proxy.load.raw_verbose("", rpc_metafile, *extra_cmds)
        for _ in range(0, 10):
            try:
                remote_proxy.d.hash(self.hash)
                break
            except rpc.HashNotFound:
                time.sleep(0.5)
        # After 5 seconds, let the exception happen
        remote_proxy.d.hash(self.hash)

        # Keep custom values. Trying to load these in during the
        # load.raw tends to cause either the load to fail or the
        # values to get corrupted, even for simple values.
        for k, v in self.custom_items().items():
            remote_proxy.d.custom.set(self.hash, k, v)
        for key in range(1, 5):
            value = getattr(proxy.d, f"custom{key}")(self.hash)
            if value:
                getattr(remote_proxy.d, f"custom{key}.set")(self.hash, value)

        remote_proxy.d.start(self.hash)
        if not copy:
            proxy.d.erase(self.hash)
            self._fields.clear()
        return True

    def delete(self):
        """Remove torrent from client."""
        self.stop()
        tied_file = self.rpc_call("d.tied_to_file")
        if tied_file:
            self._make_it_so("removing metafile of", ["d.delete_tied"])
        self._make_it_so("erasing", ["d.erase"])

    # TODO: def set_files_priority(self, pattern, prio)
    # Set priority of selected files
    # NOTE: need to call d.update_priorities after f.priority.set!

    def purge(self):
        """Delete PARTIAL data files and remove torrent from client."""

        def partial_file(item):
            "Filter out partial files"
            return item.completed_chunks < item.size_chunks

        self.cull(file_filter=partial_file, attrs=["completed_chunks", "size_chunks"])

    def cull(self, file_filter: Optional[Callable] = None, attrs: List[str] = None):
        """Delete ALL data files and remove torrent from client.

        @param file_filter: Optional callable for selecting a subset of all files.
            The callable gets a file item as described for RtorrentItem._get_files
            and must return True for items eligible for deletion.
        @param attrs: Optional list of additional attributes to fetch (for
            file_filter to use).
        """
        dry_run = False  # set to True for testing
        path = Path(self.rpc_call("d.directory"))
        if not path.is_absolute():
            raise error.EngineError(
                f"Directory '{path}' for item {self.hash} is not absolute, which is a bad idea,"
                " fix your .rtorrent.rc to use 'directory.default.set = /...'"
            )

        if not path.exists():
            return

        dirs_to_clean_up = set([path])
        for file_data in self._get_files(attrs=attrs):
            if file_filter is not None and not file_filter(file_data):
                continue
            fullpath = Path(path, file_data.path)
            if fullpath.is_file() or fullpath.is_symlink():
                self._engine.logger.debug("Deleting '%s'", fullpath)
                dirs_to_clean_up.add(fullpath.parent)
                if not dry_run:
                    fullpath.unlink()
        for directory in dirs_to_clean_up:
            try:
                if not dry_run:
                    directory.rmdir()
            except OSError as e:
                if e.errno != errno.ENOTEMPTY:
                    raise

        # Delete item from engine
        if not dry_run:
            self.delete()

    def flush(self):
        """Write volatile data to disk."""
        self._make_it_so("saving session data of", ["d.save_resume"])


class RtorrentEngine:
    """The rTorrent backend proxy."""

    # Bare minimum fields to prefetch
    BASE_FIELDS = {
        "d.hash",
    }

    # rTorrent names of fields that never change
    CONSTANT_FIELDS = BASE_FIELDS | {
        "d.name",
        "d.is_private",
        "d.is_multi_file",
        "d.tracker_size",
        "d.size_bytes",
    }

    # rTorrent names of fields that need to be pre-fetched
    CORE_FIELDS = CONSTANT_FIELDS | {
        "d.complete",
        "d.tied_to_file",
    }

    # rTorrent names of fields that get fetched in multi-call
    PREFETCH_FIELDS = CORE_FIELDS | {
        "d.base_path",
        "d.custom=memo_alias",
        "d.custom=tm_completed",
        "d.custom=tm_loaded",
        "d.custom=tm_started",
        "d.down.rate",
        "d.down.total",
        "d.is_active",
        "d.is_open",
        "d.message",
        "d.ratio",
        "d.up.rate",
        "d.up.total",
    }

    def __init__(self, url=None, auto_open=False):
        """Initialize proxy."""
        self.logger = pymagic.get_class_logger(self)
        self.engine_id = "N/A"  # ID of the instance we're connecting to
        self.engine_software = "rTorrent"  # Name and version of software
        self.startup = time.time()
        self.rpc = None
        self.properties = {}
        self.known_throttle_names = {"", "NULL"}
        self.url: str
        if url is None:
            config.autoload_scgi_url()
            self.url = config.settings.SCGI_URL
        else:
            self.url = url
        if auto_open:
            self.open()

    def view(self, viewname="default", matcher=None):
        """Get list of download items."""
        return engine.TorrentView(self, viewname, matcher)

    def __repr__(self):
        """Return a representation of internal state."""
        if self.rpc:
            # Connected state
            return "{} connected to {} [{}, up {}] via {!r}".format(
                self.__class__.__name__,
                self.engine_id,
                self.engine_software,
                fmt.human_duration(self.uptime, 0, 2, True).strip(),
                self.url,
            )
        # Unconnected state
        return "{} connectable via {!r}".format(
            self.__class__.__name__,
            self.url,
        )

    @property
    def uptime(self):
        """rTorrent's uptime."""
        return time.time() - self.startup

    def _resolve_viewname(self, viewname: str) -> str:
        """Check for special view names and return existing rTorrent one."""
        if viewname == "-":
            try:
                viewname = self.open().ui.current_view()
            except rpc.ERRORS as exc:
                raise error.EngineError(f"Can't get name of current view: {exc}")

        return viewname

    def system_multicall(self, methods: Dict[str, List]) -> Dict:
        """Helper method for system.multicall"""
        results = {}
        call = []
        for method, params in methods.items():
            call.append({"methodName": method, "params": params})
        for index, r in enumerate(self.rpc.system.multicall(call)):
            results[list(methods.keys())[index]] = r[0]
        return results

    @lru_cache(maxsize=32)
    def has_method(self, method_name: str):
        """Cached check to see if method exists"""
        if method_name in self.rpc.system.listMethods():
            return True
        return False

    def open(self):
        """Open connection."""
        # Only connect once
        if self.rpc is not None:
            return self.rpc

        # Reading abilities are on the downfall, so...
        if not self.url:
            raise error.UserError(
                "You need to configure a RPC connection, read"
                " https://kannibalox.github.io/pyrosimple/configuration/#top-level-section"
            )

        # Connect and get instance ID (also ensures we're connectable)
        self.rpc = rpc.RTorrentProxy(self.url)
        self.properties = self.system_multicall(
            {
                "system.client_version": [],
                "system.library_version": [],
                "system.time_usec": [],
                "session.name": [],
                "directory.default": [],
                "session.path": [],
            }
        )
        self.engine_id = self.properties["session.name"]
        time_usec = self.properties["system.time_usec"]

        # Make sure xmlrpc-c works as expected
        if time_usec < 2**32:
            self.logger.warning(
                "Unsupported xmlrpc-c version (64 bit integer support missing,"
                " %r returned instead)",
                type(time_usec),
            )

        # Get other manifest values
        self.engine_software = f"rTorrent {self.properties['system.library_version']}/{self.properties['system.client_version']}"

        try:
            self.startup = int(self.rpc.startup_time() or time.time())
        except xmlrpclib.Fault:
            self.startup = int(time.time())

        # Return connection
        self.logger.debug("%s", repr(self))
        return self.rpc

    def multicall(self, viewname: str, fields: List[str]) -> List[Bunch]:
        """Query the given fields of items in the given view.

        The result list contains named tuples,
        so you can access the fields directly by their name.
        """
        commands = tuple(f"d.{x}=" for x in fields)
        items = self.open().d.multicall2("", viewname, *commands)
        return [
            Bunch(dict(zip([x.replace(".", "_") for x in fields], item)))
            for item in items
        ]

    def log(self, msg: str):
        """Log a message in the torrent client."""
        self.open().log(rpc.NOHASH, msg)

    def item(self, infohash: str, prefetch=None):
        """Fetch a single item by its info hash."""
        return next(self.items(infohash, prefetch))

    def items(
        self,
        view: Optional[Union[engine.TorrentView, str]] = None,
        prefetch: Optional[Set[str]] = None,
    ):
        """Get list of download items.

        @param view: Name of the view.
        @param prefetch: Optional list of field names to fetch initially.
        """

        if view is None:
            view = engine.TorrentView(self, "default")
        elif isinstance(view, str):
            view = engine.TorrentView(self, self._resolve_viewname(view))

        # Map pyroscope names to rTorrent ones
        if prefetch:
            prefetch = self.BASE_FIELDS | set(prefetch)
        else:
            prefetch = self.PREFETCH_FIELDS

        # Fetch items
        items = []
        multi_args: List
        try:
            # Prepare multi-call arguments
            args = sorted(prefetch)

            # Check if view is in the format of a single hash
            infohash = view._check_hash_view()
            if infohash:
                multi_call = self.open().system.multicall
                multi_args = [
                    dict(
                        methodName=field.rsplit("=", 1)[0],
                        params=[infohash]
                        + (field.rsplit("=", 1)[1].split(",") if "=" in field else []),
                    )
                    for field in args
                ]
                raw_items = [[i[0] for i in multi_call(multi_args)]]
            # Otherwise prepare a multicall as expected
            else:
                multi_call = self.open().d.multicall2
                multi_args = ["", view.viewname] + [
                    field if "=" in field else field + "=" for field in args
                ]
                if view.matcher and int(config.settings.get("FAST_QUERY")) > 0:
                    pre_filter = ""
                    if config.settings.SAFETY_CHECKS_ENABLED and not self.has_method(
                        "d.multicall.filtered"
                    ):
                        self.logger.warning(
                            "Fast query enabled but host does not support 'd.multicall.filtered', disabling fast query."
                        )
                    else:
                        pre_filter = matching.unquote_pre_filter(
                            view.matcher.pre_filter()
                        )
                    if pre_filter:
                        # rTorrent 0.9.8+ does not have
                        # sting.contains_i, so we check for it here
                        self.logger.debug("Created pre-filter: %s", pre_filter or "N/A")
                        if (
                            config.settings.SAFETY_CHECKS_ENABLED
                            and "string.contains_i" in pre_filter
                            and not self.has_method("string.contains_i")
                        ):
                            self.logger.warning(
                                "Method 'strings.contains_i' does not exist!"
                                "Fast query %r would return an empty list, disabling fast query.",
                                pre_filter,
                            )
                        else:
                            multi_call = self.open().d.multicall.filtered
                            multi_args.insert(2, pre_filter)
                raw_items = multi_call(*tuple(multi_args))

            self.logger.debug(
                "Got %d items with %d attributes",
                len(raw_items),
                len(prefetch),
            )

            # Build objects from the received data
            for item in raw_items:
                ritem = RtorrentItem(
                    self,
                    fields={},
                    rpc_fields=dict(zip(args, item)),
                )

                if view.matcher:
                    if view.matcher.match(ritem):
                        items.append(ritem)
                        yield items[-1]
                else:
                    items.append(ritem)
                    yield items[-1]
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting download items from {self!r}: {exc}"
            )

    def show(
        self,
        items,
        view: Optional[str] = None,
        append: bool = False,
        disjoin: bool = False,
    ):
        """Place a set of items (search result) into a view, and
        return the view name."""
        proxy = self.open()
        view_name: str = self._resolve_viewname(view or "rtcontrol")

        if append and disjoin:
            raise error.EngineError(
                f"Cannot BOTH append to / disjoin from view '{view_name}'"
            )

        # Add view if needed
        if view not in proxy.view.list():
            proxy.view.add(rpc.NOHASH, view_name)

        # Clear view and show it
        if not append and not disjoin:
            proxy.view.filter(rpc.NOHASH, view_name, "false=")
            proxy.d.multicall2(rpc.NOHASH, "default", "d.views.remove=" + view_name)
        proxy.ui.current_view.set(rpc.NOHASH, view_name)

        # Add items
        for item in items:
            if disjoin:
                proxy.d.views.remove(item.hash, view_name)
                proxy.view.set_not_visible(item.hash, view_name)
            else:
                proxy.d.views.push_back_unique(item.hash, view_name)
                proxy.view.set_visible(item.hash, view_name)

        return view


def expand_template(template_path: str, namespace: Dict) -> str:
    """Expand the given template file.
    Currently, only jinja2 templates are supported.

    @param template: The name of the template, to be loaded by the jinja2 loaders.
    @param namespace: Custom namespace that is added to the predefined defaults
        and takes precedence over those.
    @return: The expanded template.
    @raise LoggableError: In case of typical errors during template execution.
    """
    template = env.get_template(template_path)
    # Default templating namespace
    # variables = dict(c=config.custom_template_helpers)
    variables = {}
    # Provided namespace takes precedence
    variables.update(namespace)
    # Expand template
    return template.render(**variables)


def format_item_str(
    template_str: str, item: Union[Dict, str, RtorrentItem], defaults=None
):
    """Simple helper function to format a string with an item"""
    template = env.from_string(template_str)
    return format_item(template, item, defaults)


def format_item(
    template: jinja2.Template,
    item: Union[Dict, str, RtorrentItem],
    defaults: Optional[Dict] = None,
) -> str:
    """Format an item according to the given output template.

    @param format_spec: The output template, preparsed by jinja2.
    @param item: The object, which is automatically wrapped for interpolation.
    @param defaults: Optional default values.
    """
    if defaults is None:
        defaults = {}
    return str(template.render(d=item, **defaults))


def validate_field_list(
    fields: str,
    allow_fmt_specs: bool = False,
) -> List[str]:
    """Make sure the fields in the given list exist.

    @param fields: List of fields (comma-/space-separated if a string).
    @type fields: str
    @return: validated field names.
    @rtype: list
    """
    split_fields = [i.strip() for i in fields.split(",")]

    for name in split_fields:
        if allow_fmt_specs and "." in name:
            fullname = name
            name, fmtspecs = name.split(".", 1)
            for fmtspec in fmtspecs.split("."):
                if fmtspec not in env.filters and fmtspec != "raw":
                    raise error.UserError(
                        f"Unknown format specification {fmtspec!r} in {fullname!r}"
                    )
        if (
            name not in engine.FIELD_REGISTRY
            and not engine.TorrentProxy.add_manifold_attribute(name)
        ):
            raise error.UserError(f"Unknown field name {name!r}")

    return split_fields


def validate_sort_fields(sort_fields: str):
    """Make sure the fields in the given list exist, and return sorting key.

    If field names are prefixed with '-', sort order is reversed for that field (descending).
    """
    # Create sort specification
    sort_spec: Tuple = tuple()
    for name in sort_fields.split(","):
        name = name.strip()
        descending = False
        if name.startswith("-"):
            name = name[1:]
            descending = True
        sort_spec += ((name, descending),)

    # Validate field list
    validate_field_list(",".join([name for name, _ in sort_spec]))
    logger = logging.getLogger(__name__)
    logger.debug(
        "Validated key names: %s",
        ", ".join([("-" if descending else "") + i for i, descending in sort_spec]),
    )

    # Need to provide complex key in order to allow for the minimum
    # amount of attribute fetches, since they could mean a potentially
    # expensive RPC call.
    class Key:
        "Complex sort order key"

        def __init__(self, obj, *_):
            "Remember object to be compared"
            self.obj = obj

        def __lt__(self, other):
            "Compare to other key"
            for field, descending in sort_spec:
                lhs, rhs = getattr(self.obj, field), getattr(other.obj, field)
                if lhs == rhs:
                    continue
                return rhs < lhs if descending else lhs < rhs
            return False

    return Key


def get_fields_from_template(
    template: str, item_name: str = "d"
) -> Generator[str, None, None]:
    """Utility function to get field references from a template

    E.g: 'Size: {{d.size}}' -> ['size']"""
    for node in env.parse(template).find_all(jinja2.nodes.Getattr):
        if isinstance(node.node, jinja2.nodes.Name) and node.node.name == item_name:
            yield node.attr



"""Handles RPC methods over various transports"""
import io
import logging
import socket
import subprocess
import sys

from typing import Dict, List, Tuple, Type
from urllib import parse as urlparse
from urllib.error import URLError
from xmlrpc import client as xmlrpclib

from prometheus_client import Counter, Summary


request_counter = Counter(
    "transport_request", "Number of requests made by the transport"
)
request_size_counter = Counter("transport_request_size", "Size of the request in bytes")
response_time_summary = Summary("response_time", "Time spent watiing for a response")
response_size_counter = Counter(
    "transport_response_size", "Size of the response in bytes"
)

logger = logging.getLogger(__name__)


class SCGIException(Exception):
    """SCGI protocol error"""


# Types of exceptions thrown
ERRORS = (SCGIException, URLError, xmlrpclib.Fault, socket.error)


class RTorrentTransport(xmlrpclib.Transport):
    """Base class for handling transports. Primarily exists to allow
    using the different transports with different underlying RPC mechanisms"""

    def __init__(self, *args, url, codec=xmlrpclib, headers=(), **kwargs):
        if "/" not in url and ":" in url and url.rsplit(":")[-1].isdigit():
            url = "scgi://" + url
        if url.startswith("/") or url.startswith("~"):
            url = "scgi+unix://" + url
        self.url = url
        self.codec = codec
        self.verbose = False
        # We need to handle the headers differently based on the RPC protocols
        self._headers = list(headers)
        # Pass them to the transport in py 3.8+ only
        if sys.version_info.minor >= 8:
            kwargs["headers"] = self._headers
        super().__init__(*args, **kwargs)

    def parse_response(self, response):
        if self.codec == xmlrpclib:
            return super().parse_response(response)
        return self.codec.loads(response.read())


class SSHTransport(RTorrentTransport):
    """Transport via SSH conneection."""

    def request(self, host, handler, request_body, verbose=False):
        request_counter.inc()
        request_size_counter.inc(len(request_body))
        self.verbose = verbose
        target = urlparse.urlparse(self.url).path
        cmd = ["ssh", host, "socat", "STDIO", target[1:]]
        with response_time_summary.time():
            resp = subprocess.run(
                cmd,
                input=_encode_payload(request_body, self._headers),
                capture_output=True,
                check=False,
            )
        if resp.returncode > 0:
            logger.error("SSH command returned non-zero exit code")
            logger.error("stderr: %s", resp.stderr)
            logger.error("stdout: %s", resp.stdout)
        response_size_counter.inc(len(resp.stdout))
        return self.parse_response(io.BytesIO(_parse_response(resp.stdout)[0]))


class HTTPTransport(RTorrentTransport):
    """Transport via HTTP(s) call."""

    # Notably the request here is *not* encoded into SCGI
    # since the web proxy handles that itself.
    def request(self, host, handler, request_body, verbose=False):
        # Defer loading for performance reasons
        import requests  # pylint: disable=import-outside-toplevel

        request_counter.inc()
        request_size_counter.inc(len(request_body))
        headers = {k.replace("_", "-"): v for k, v in dict(self._headers).items()}
        with response_time_summary.time():
            req = requests.post(
                self.url, headers=headers, data=request_body, timeout=60
            )
        response_size_counter.inc(len(req.content))
        req.raise_for_status()
        return self.parse_response(io.BytesIO(req.content))


class TCPTransport(RTorrentTransport):
    """Transport via TCP socket."""

    def request(self, host, handler, request_body, verbose=False):
        request_counter.inc()
        request_size_counter.inc(len(request_body))
        self.verbose = verbose
        target = urlparse.urlparse(self.url)
        with response_time_summary.time():
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                host, port = target.netloc.split(":")
                sock.connect((host, int(port)))
                sock.sendall(_encode_payload(request_body, self._headers))
                with sock.makefile("rb") as handle:
                    response = _parse_response(handle.read())[0]
        response_size_counter.inc(len(response))
        return self.parse_response(io.BytesIO(response))


class UnixTransport(RTorrentTransport):
    """Transport via UNIX domain socket."""

    def request(self, _host, handler, request_body, verbose=False):
        request_counter.inc()
        request_size_counter.inc(len(request_body))
        self.verbose = verbose
        target = urlparse.urlparse(self.url).path
        with response_time_summary.time():
            with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
                sock.connect(target)
                sock.sendall(_encode_payload(request_body))
                with sock.makefile("b") as handle:
                    response = _parse_response(handle.read())[0]
        response_size_counter.inc(len(response))
        return self.parse_response(io.BytesIO(response))


TRANSPORTS = {
    "scgi": TCPTransport,
    "http": HTTPTransport,
    "https": HTTPTransport,
    "scgi+unix": UnixTransport,
    "scgi+ssh": SSHTransport,
}


def register_scheme(scheme: str):
    """Helper method to register protocols with urllib"""
    for method in filter(lambda s: s.startswith("uses_"), dir(urlparse)):
        getattr(urlparse, method).append(scheme)


for t in TRANSPORTS:
    register_scheme(t)


def transport_from_url(url: str) -> Type[xmlrpclib.Transport]:
    """Create a transport for the given URL."""
    if "/" not in url and ":" in url and url.rsplit(":")[-1].isdigit():
        url = "scgi://" + url
    if url.startswith("/") or url.startswith("~"):
        url = "scgi+unix://" + url
    parsed_url = urlparse.urlsplit(url, allow_fragments=False)
    try:
        transport = TRANSPORTS[parsed_url.scheme.lower()]
        return transport
    except KeyError:
        # Support simplified "domain:port" URLs
        if not any((parsed_url.netloc, parsed_url.query)) and parsed_url.path.isdigit():
            return transport_from_url(f"scgi://{parsed_url.netloc}:{parsed_url.path}")
    raise URLError(f"Unsupported scheme in URL {parsed_url.geturl()!r}")


def _encode_netstring(data: bytes) -> bytes:
    "Encode data as netstring."
    return b"%d:%s," % (len(data), data)


def _encode_headers(headers: List[Tuple[str, str]]) -> bytes:
    "Make SCGI header bytes from list of tuples."
    return b"".join(
        [b"%s\0%s\0" % (k.encode("ascii"), v.encode("ascii")) for k, v in headers]
    )


def _encode_payload(data: bytes, headers: List[Tuple[str, str]] = None) -> bytes:
    "Wrap data in an SCGI request."
    prolog: bytes = b"CONTENT_LENGTH\0%d\0SCGI\x001\0" % len(data)
    if headers:
        prolog += _encode_headers(headers)

    return _encode_netstring(prolog) + data


def _parse_headers(headers: bytes) -> Dict[str, str]:
    """
    Get headers dict from header bytestring.

    :param headers: An SCGI header bytestring
    :type headers: bytes
    :return: A dictionary of string keys/values
    """
    try:
        result = {}
        for line in headers.splitlines():
            if line:
                key, value = line.rstrip().split(b": ", 1)
                result[key.decode("ascii")] = value.decode("ascii")
        return result
    except (TypeError, ValueError) as exc:
        raise SCGIException(f"Error in SCGI headers {headers.decode()}") from exc


def _parse_response(resp: bytes) -> Tuple[bytes, Dict[str, str]]:
    """
    Get xmlrpc response from scgi response

    :param headers: An SCGI bytestring payload
    :type headers: bytes
    :return: A tuple of a binary payload and a dictionary of string keys/values
    """

    # Assume they care for standards and send us CRLF (not just LF)
    try:
        headers, payload = resp.split(b"\r\n\r\n", 1)
    except (TypeError, ValueError) as exc:
        raise SCGIException(
            f"No header delimiter in SCGI response of length {len(resp)}"
        ) from exc
    parsed_headers = _parse_headers(headers)

    clen = parsed_headers.get("Content-Length")
    if clen is not None:
        # Check length, just in case the transport is bogus
        assert len(payload) == int(clen)

    return payload, parsed_headers



""" Metafile Editor.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import copy
import difflib
import hashlib
import logging
import os
import re
import sys
import time
import urllib.parse

from pathlib import Path
from typing import List, Optional, cast

import bencode  # typing: ignore

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBase
from pyrosimple.util import metafile


def replace_fields(meta, patterns):
    """Replace patterns in fields."""
    for pattern in patterns:
        try:
            field, regex, subst, _ = pattern.split(pattern[-1])

            # TODO: Allow numerical indices, and "+" for append
            namespace = meta
            keypath = [
                i.replace("\0", ".") for i in field.replace("..", "\0").split(".")
            ]
            for key in keypath[:-1]:
                namespace = namespace[key]

            namespace[keypath[-1]] = re.sub(regex, subst, namespace[keypath[-1]])
        except (KeyError, IndexError, TypeError, ValueError) as exc:
            raise error.UserError(f"Bad substitution '{pattern}' ({exc})!")

    return meta


def diff_metafiles(
    meta1: metafile.Metafile,
    meta2: metafile.Metafile,
    meta1_name="old",
    meta2_name="new",
) -> List[str]:
    """Return a git-style diff between two metafiles"""

    from pyrosimple.util.fmt import (  # pylint: disable=import-outside-toplevel
        BencodeJSONEncoder,
    )

    def encode_meta(meta: metafile.Metafile) -> List[str]:
        """Sanitize and modify meta for better diffing"""
        meta_dict = meta.dict_copy()
        if "info" in meta_dict and "pieces" in meta_dict["info"]:
            meta_dict["info"][
                "pieces"
            ] = f'{len(meta_dict["info"]["pieces"])//20} pieces, md5 hash {hashlib.md5(meta_dict["info"]["pieces"]).hexdigest()}'

        return [
            l + "\n" for l in BencodeJSONEncoder(indent=2).encode(meta_dict).split("\n")
        ]

    meta1_out = encode_meta(meta1)
    meta2_out = encode_meta(meta2)
    return list(difflib.unified_diff(meta1_out, meta2_out, meta1_name, meta2_name))


class MetafileChanger(ScriptBase):
    """Change attributes of a bittorrent metafile."""

    # Keys of rTorrent session data
    RT_RESUME_KEYS = ("libtorrent_resume", "log_callback", "err_callback", "rtorrent")

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.parser.add_argument("metafile", nargs="+", help="Torrent files to modify")
        self.add_bool_option(
            "-n",
            "--dry-run",
            help="don't write changes to disk, just tell what would happen",
        )
        self.add_bool_option(
            "--diff",
            help="display a diff for each change being made",
        )
        self.add_bool_option(
            "-V",
            "--no-skip",
            help="do not skip broken metafiles that fail the integrity check",
        )
        self.add_value_option(
            "-o",
            "--output-directory",
            "PATH",
            help="optional output directory for the modified metafile(s)",
        )
        self.add_bool_option(
            "-p", "--make-private", help="make torrent private (DHT/PEX disabled)"
        )
        self.add_bool_option(
            "-P", "--make-public", help="make torrent public (DHT/PEX enabled)"
        )
        self.add_value_option(
            "-s",
            "--set",
            "KEY=VAL",
            action="append",
            default=[],
            help="set a specific key to the given value; omit the '=' to delete a key",
        )
        self.add_value_option(
            "-r",
            "--regex",
            "KEY/REGEX/SUBST/",
            action="append",
            default=[],
            help="replace pattern in a specific key by the given substitution",
        )
        self.add_bool_option(
            "-C",
            "--clean",
            help="remove all non-standard data from metafile outside the info dict",
        )
        self.add_bool_option(
            "-A",
            "--clean-all",
            help="remove all non-standard data from metafile including inside the info dict",
        )
        self.add_bool_option(
            "-X",
            "--clean-xseed",
            help="like --clean-all, but keep libtorrent resume information",
        )
        self.add_bool_option(
            "-R",
            "--clean-rtorrent",
            help="remove all rTorrent session data from metafile",
        )
        self.add_value_option(
            "-H",
            "--hashed",
            "--fast-resume",
            "DATAPATH",
            help="add libtorrent fast-resume information (use {} in place of the torrent's name in DATAPATH)",
        )
        self.add_value_option(
            "-c",
            "--check-data",
            "PATH",
            help="check the hash against the data in the given path before making any changes",
        )
        self.add_value_option(
            "-T",
            "--tracker",
            "DOMAIN",
            help="filter given torrents for a tracker domain",
        )
        self.add_value_option(
            "-a",
            "--reannounce",
            "URL",
            help="set a new announce URL, but only if the old announce URL matches the new one",
        )
        self.add_value_option(
            "--reannounce-all",
            "URL",
            help="set a new announce URL on ALL given metafiles",
        )
        self.add_bool_option(
            "--no-cross-seed",
            help="when using --reannounce-all, do not add a non-standard field to the info dict ensuring unique info hashes",
        )
        self.add_value_option(
            "--comment", "TEXT", help="set a new comment (an empty value deletes it)"
        )
        self.add_bool_option("--bump-date", help="set the creation date to right now")
        self.add_bool_option("--no-date", help="remove the 'creation date' field")

    def mainloop(self) -> None:
        """The main loop."""
        self.args = self.options.metafile
        if not self.args:
            self.parser.print_help()
            self.parser.error("No metafiles given, nothing to do!")
            self.parser.exit()

        if self.options.reannounce and self.options.reannounce_all:
            self.parser.error("Conflicting options --reannounce and --reannounce-all!")

        # Set filter criteria for metafiles
        filter_url_prefix: Optional[str] = None
        if self.options.reannounce:
            filter_url_prefix = config.map_announce2alias(self.options.reannounce)
            self.log.info(
                "Filtering for metafiles with announce URL %r...",
                filter_url_prefix,
            )
        elif self.options.tracker:
            filter_url_prefix = config.map_announce2alias(self.options.tracker)
            self.log.info(
                "Filtering for metafiles with tracker %r...",
                filter_url_prefix,
            )
        if self.options.reannounce_all:
            self.options.reannounce = self.options.reannounce_all
        else:
            # When changing the announce URL w/o changing the domain, don't change the info hash!
            self.options.no_cross_seed = True

        # Resolve tracker alias, if URL doesn't look like an URL
        if (
            self.options.reannounce
            and not urllib.parse.urlparse(self.options.reannounce).scheme
        ):
            for conn in config.multi_connection_lookup(self.options.reannounce):
                if urllib.parse.urlparse(conn).scheme:
                    self.options.reannounce = conn
                    break
            else:
                raise error.UserError(
                    "Unknown tracker alias or URL %r!" % (self.options.reannounce)
                )

        # go through given files
        bad = 0
        changed = 0
        for filename in self.options.metafile:
            try:
                # Read and remember current content
                torrent: metafile.Metafile = metafile.Metafile.from_file(Path(filename))
            except (OSError, KeyError, bencode.BencodeDecodeError) as exc:
                self.log.warning(
                    "Skipping bad metafile %r (%s: %s)",
                    filename,
                    type(exc).__name__,
                    exc,
                )
                bad += 1
            else:
                # Check metafile integrity
                try:
                    torrent.check_meta()
                    if self.options.check_data:
                        # pylint: disable=import-outside-toplevel
                        from pyrosimple.util.metafile import PieceLogger
                        from pyrosimple.util.ui import (
                            HashProgressBar,
                            HashProgressBarCounter,
                        )

                        with HashProgressBar() as pb:
                            if (
                                logging.getLogger().isEnabledFor(logging.WARNING)
                                and sys.stdout.isatty()
                            ):
                                progress_callback = cast(
                                    HashProgressBarCounter, pb()
                                ).progress_callback
                            else:
                                progress_callback = None

                            piece_logger = PieceLogger(torrent, self.log)

                            data_correct = torrent.hash_check(
                                Path(self.options.check_data),
                                progress_callback=progress_callback,
                                piece_callback=piece_logger.check_piece,
                            )
                        if not data_correct:
                            self.log.error(
                                "File %s does match data from %s",
                                filename,
                                self.options.check_data,
                            )
                            sys.exit(error.EX_DATAERR)
                except ValueError as exc:
                    self.log.warning(
                        "Metafile %r failed integrity check: %s",
                        filename,
                        exc,
                    )
                    if not self.options.no_skip:
                        continue

                # Skip any metafiles that don't meet the pre-conditions
                if (
                    filter_url_prefix
                    and config.map_announce2alias(torrent["announce"])
                    != filter_url_prefix
                ):
                    self.log.info(
                        "Skipping metafile %r: not tracked by %r!",
                        filename,
                        filter_url_prefix,
                    )
                    continue

                # Keep resume info safe
                libtorrent_resume = {}
                if "libtorrent_resume" in torrent:
                    try:
                        libtorrent_resume["bitfield"] = torrent["libtorrent_resume"][
                            "bitfield"
                        ]
                    except KeyError:
                        pass  # nothing to remember

                    libtorrent_resume["files"] = copy.deepcopy(
                        torrent["libtorrent_resume"]["files"]
                    )

                # Change private flag?
                if self.options.make_private and not torrent["info"].get("private", 0):
                    self.log.info("Setting private flag...")
                    torrent["info"]["private"] = 1
                if self.options.make_public and torrent["info"].get("private", 0):
                    self.log.info("Clearing private flag...")
                    del torrent["info"]["private"]

                # Remove non-standard keys?
                if (
                    self.options.clean
                    or self.options.clean_all
                    or self.options.clean_xseed
                ):
                    torrent.clean_meta(
                        including_info=not self.options.clean,
                    )

                # Restore resume info?
                if self.options.clean_xseed:
                    if libtorrent_resume:
                        self.log.info("Restoring key 'libtorrent_resume'...")
                        torrent.setdefault("libtorrent_resume", {})
                        torrent["libtorrent_resume"].update(libtorrent_resume)
                    else:
                        self.log.warning("No resume information found!")

                # Clean rTorrent data?
                if self.options.clean_rtorrent:
                    for key in self.RT_RESUME_KEYS:
                        if key in torrent:
                            self.log.info("Removing key %r...", key)
                            del torrent[key]

                # Change announce URL?
                if self.options.reannounce:
                    torrent["announce"] = self.options.reannounce
                    if "announce-list" in torrent:
                        del torrent["announce-list"]

                    if not self.options.no_cross_seed:
                        # Enforce unique hash per tracker
                        torrent["info"]["x_cross_seed"] = hashlib.md5(
                            self.options.reannounce.encode()
                        ).hexdigest()

                # Change comment or creation date?
                if self.options.comment is not None:
                    if self.options.comment:
                        torrent["comment"] = self.options.comment
                    elif "comment" in torrent:
                        del torrent["comment"]
                if self.options.bump_date:
                    torrent["creation date"] = int(time.time())
                if self.options.no_date and "creation date" in torrent:
                    del torrent["creation date"]

                # Add fast-resume data?
                if self.options.hashed:
                    datadir: str = self.options.hashed
                    if "{}" in datadir and not os.path.exists(datadir):
                        datadir = datadir.replace("{}", torrent["info"]["name"])
                    try:
                        torrent.add_fast_resume(Path(datadir))
                    except OSError as exc:
                        self.fatal("Error making fast-resume data (%s)", exc)
                        raise

                # Set specific keys?
                torrent.assign_fields(self.options.set)
                replace_fields(torrent, self.options.regex)

                # Write new metafile, if changed

                if self.options.output_directory:
                    filename = os.path.join(
                        self.options.output_directory, os.path.basename(filename)
                    )
                    self.log.info("Will write %r...", filename)

                    if not self.options.dry_run:
                        with open(filename, "wb") as fh:
                            fh.write(bencode.bencode(torrent))
                        if "libtorrent_resume" in torrent:
                            # Also write clean version
                            filename = filename.replace(
                                ".torrent", "-no-resume.torrent"
                            )
                            del torrent["libtorrent_resume"]
                            self.log.info("Writing '%s'...", filename)
                            bencode.bwrite(torrent, filename)
                else:
                    current_torrent = metafile.Metafile.from_file(filename)
                    diff = diff_metafiles(current_torrent, torrent, filename)
                    if self.options.diff:
                        sys.stdout.writelines(diff)
                    if not diff:
                        self.log.info("No change to file %r", filename)
                        continue
                    self.log.info("Changing %r...", filename)

                    if not self.options.dry_run:
                        # Write to temporary file
                        tempname = os.path.join(
                            os.path.dirname(filename),
                            "." + os.path.basename(filename),
                        )
                        self.log.debug("Writing temporary file '%s'...", tempname)
                        torrent.save(Path(tempname))

                        try:
                            self.log.debug("Replacing file '%s'...", filename)
                            os.replace(tempname, filename)
                        except OSError as exc:
                            # TODO: Try to write directly, keeping a backup!
                            raise error.LoggableError(
                                "Can't rename tempfile %r to %r (%s)"
                                % (tempname, filename, exc)
                            )

                changed += 1

        # Print summary
        if changed:
            self.log.info(
                "%s %d metafile(s).",
                "Would've changed" if self.options.dry_run else "Changed",
                changed,
            )
        if bad:
            self.log.warning("Skipped %d bad metafile(s)!", bad)


def run():  # pragma: no cover
    """The entry point."""
    MetafileChanger().run()


if __name__ == "__main__":
    run()



"""A simple job to emulate rtcontrol's actions"""
import subprocess

from pyrosimple import error
from pyrosimple.job.base import BaseJob, MatchableJob
from pyrosimple.torrent import rtorrent


class Command(BaseJob):
    """Runs a single untemplated command."""

    def __init__(self, config=None):
        super().__init__(config)
        self.args = self.config.get("args", [])
        allowed_kwargs = ["shell", "cwd", "timeout", "check", "env"]
        self.kwargs = {k: self.config[k] for k in allowed_kwargs if k in self.config}

    def run(self):
        if not self.config["dry_run"]:
            proc = subprocess.run(  # pylint: disable=subprocess-run-check
                self.args, **self.kwargs, capture_output=True
            )
            self.log.info("Command %s finished with RC=%s", proc.args, proc.returncode)
            self.log.debug("stdout: %s", proc.stdout)
            self.log.debug("stderr: %s", proc.stderr)
        else:
            self.log.info("Would run %s with parameters %s", self.args, self.kwargs)


class ItemCommand(MatchableJob):
    """Runs a templated command against matching items."""

    def __init__(self, config=None):
        super().__init__(config)
        self.args = [rtorrent.env.from_string(a) for a in self.config.get("args", [])]
        allowed_kwargs = ["shell", "cwd", "timeout", "check", "env"]
        self.kwargs = {k: self.config[k] for k in allowed_kwargs if k in self.config}

    def run_item(self, item):
        if not self.config["dry_run"]:
            proc = subprocess.run(  # pylint: disable=subprocess-run-check
                [rtorrent.format_item(a, item) for a in self.args],
                **self.kwargs,
                capture_output=True,
            )
            self.log.info("Command %s finished with RC=%s", proc.args, proc.returncode)
            self.log.debug("stdout: %s", proc.stdout)
            self.log.debug("stderr: %s", proc.stderr)
        else:
            self.log.info("Would run %s with parameters %s", self.args, self.kwargs)


class Action(MatchableJob):
    """Run an action for each matched item"""

    def __init__(self, config=None):
        super().__init__(config)
        self.action = self.config["action"]
        if not hasattr(rtorrent.RtorrentItem, self.action):
            raise error.ConfigurationError(f"Action '{self.action}' not found!")
        self.args = [rtorrent.env.from_string(a) for a in self.config.get("args", [])]

    def run_item(self, item):
        """For now, simply call the named methods on the item"""
        action = self.config["action"]
        processed_args = [rtorrent.format_item(a, item) for a in self.args]
        if self.config["dry_run"]:
            self.log.info(
                "Would %s(%s) %s",
                action,
                "" if not processed_args else processed_args,
                item.hash,
            )
        else:
            getattr(item, self.action)(*processed_args)



"""Torrent Item Filters.

Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>

There's a lot of magic going on in this module, but essentially
its primary responsibility is parsing queries from rtcontrol and other
places. Once parsed, it uses the visitor pattern in some classes to do
things like return lists of the fields being referenced, prepare
d.multicall.filtered statements or check if it matches against an
actual item.
"""


import fnmatch
import math
import operator
import re
import time

from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Sequence, Union

from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor

from pyrosimple import config, error, torrent


TRUE = {
    "true",
    "t",
    "yes",
    "y",
    "1",
}

FALSE = {
    "false",
    "f",
    "no",
    "n",
    "0",
}


@dataclass
class FilterOperator:
    """Small class to hold operator information"""

    name: str
    query_repr: str


Operators = {
    "eq": FilterOperator("eq", "="),
    "ne": FilterOperator("ne", "!="),
    "gt": FilterOperator("gt", ">"),
    "lt": FilterOperator("lt", "<"),
    "ge": FilterOperator("ge", ">="),
    "le": FilterOperator("le", "<="),
}


def truth(val, context="statement") -> bool:
    """Convert truth value in "val" to a boolean."""
    # Try coercing it to an int then a bool
    try:
        return bool(0 + val)
    except TypeError:
        pass

    lower_val = val.lower()

    if lower_val in TRUE:
        return True
    if lower_val in FALSE:
        return False

    raise FilterError(
        f"Bad boolean value {val!r} in {context!r} (expected one of '{TRUE}', or '{FALSE}')"
    )


def unquote_pre_filter(
    pre_filter: str, regex_: re.Pattern = re.compile(r"[\\]+")
) -> str:
    """Unquote a pre-filter condition."""
    if pre_filter.startswith('"') and pre_filter.endswith('"'):
        # Unquote outer level
        pre_filter = pre_filter[1:-1]
        pre_filter = regex_.sub(
            lambda x: x.group(0)[: len(x.group(0)) // 2], pre_filter
        )

    return pre_filter


class FilterError(error.UserError):
    """(Syntax) error in filter."""


class MatcherNode:
    """Base class for the tree structure."""

    def __init__(self, children: List):
        self.children = list(children)

    def match(self, item) -> bool:
        """Check if the item matches. All logic is deferred to subclasses."""
        raise NotImplementedError()

    def __repr__(self):
        result = type(self).__name__
        if self.children:
            result = str([str(c) for c in self.children])
        return result


class GroupNode(MatcherNode):
    """This simply groups another node, and optionally inverts it (a logical NOT)"""

    def __init__(self, children: List, invert: bool):
        super().__init__(children)
        self.invert = bool(invert)

    def match(self, item) -> bool:
        assert len(self.children) == 1
        result = bool(self.children[0].match(item))
        if self.invert:
            return not result
        return result

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        assert len(self.children) == 1
        inner = str(self.children[0].pre_filter())
        if self.invert:
            if inner.startswith('"not=$') and inner.endswith('"') and "\\" not in inner:
                return inner[6:-1]  # double negation, return inner command
            if inner.startswith('"'):
                inner = '"$' + inner[1:]
            else:
                inner = "$" + inner
            return "not=" + inner
        return inner

    def __repr__(self):
        prefix = "Not" if self.invert else ""
        return f"{prefix}{type(self).__name__}{[repr(c) for c in self.children]}"


class AndNode(MatcherNode):
    """This node performs a logical AND for all of it's children."""

    def match(self, item) -> bool:
        return all(c.match(item) for c in self.children)

    def pre_filter(self):
        """Return rTorrent condition to speed up data transfer."""
        result = [x.pre_filter() for x in self.children]
        result = [x for x in result if x]
        if result:
            if int(config.settings.get("FAST_QUERY")) == 1 or len(result) == 1:
                return result[0]  # using just one simple expression is safer
            return f'and={",".join(result)}'
        return ""

    def __repr__(self):
        return f"{type(self).__name__}{[repr(c) for c in self.children]}"


class OrNode(MatcherNode):
    """This node performs a logical OR for all of it's children."""

    def match(self, item) -> bool:
        return any(c.match(item) for c in self.children)

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        if int(config.settings.get("FAST_QUERY")) == 1:
            return ""
        if len(self.children) == 1:
            return str(self.children[0].pre_filter())
        result = [x.pre_filter() for x in self.children]
        result = [x for x in result if x]
        if result:
            return f'or={",".join(result)}'
        return ""

    def __repr__(self):
        return f"{type(self).__name__}{[repr(c) for c in self.children]}"


class FieldFilter(MatcherNode):
    """Base class for all field filters.

    Subclasses of FieldFilter act as the leaves of the tree, providing
    matching and pre-filtering functionality."""

    def __init__(self, name: str, op: FilterOperator, value: str):
        """Store field name and filter value for later evaluations."""
        super().__init__([])  # Filters are the leaves of the tree and have no children
        self._name = name
        self._condition = self._value = value
        self._op: FilterOperator = op

        self.validate()

    def __str__(self) -> str:
        return str(self._name) + self._op.query_repr + str(self._condition)

    def __call__(self, item):
        return self.match(item)

    def validate(self):
        """Validate filter condition (template method)."""

    def match(self, item) -> bool:
        """Test if item matches filter.

        By default this will defer to the operator functions in subclasses,
        but that behaivor can be overridden."""
        return bool(getattr(self, self._op.name)(item))

    def eq(self, item) -> bool:
        """Test equality against item"""
        raise FilterError(
            f"Filter '{type(self).__name__}' for field '{self._name}' does not support comparison '{self._op}'"
        )

    def gt(self, item) -> bool:
        """Test if item is greater than value"""
        raise FilterError(
            f"Filter '{type(self).__name__}' for field '{self._name}' does not support comparison '{self._op}'"
        )

    # Unfortunate but necessary boilerplate functions
    # This makes it so that by default all operators are available to fields
    # If there's a more effecient way to implement it for a specific subclass, it needs
    # to happen in there.
    # pylint: disable=missing-function-docstring
    def ge(self, item) -> bool:
        return self.eq(item) or self.gt(item)

    def ne(self, item) -> bool:
        return not self.eq(item)

    def le(self, item) -> bool:
        return self.eq(item) or not self.gt(item)

    def lt(self, item) -> bool:
        return not self.eq(item) and not self.gt(item)

    # pylint: enable=missing-function-docstring

    def pre_filter(self) -> str:
        """Create a prefilter condition (if possible).

        By default this will defer to the operator functions in subclasses,
        but that behavior can be overridden."""
        method_name = f"pre_filter_{self._op.name}"
        return str(getattr(self, method_name)())

    def pre_filter_eq(self) -> str:
        """Returns empty if not defined in subclass."""
        return ""

    def pre_filter_gt(self) -> str:
        """Returns empty if not defined."""
        return ""

    def pre_filter_ne(self) -> str:
        """Create a prefilter by creating a NOT[name=filter] object and rendering it."""
        if self.pre_filter_eq():
            return GroupNode(
                [type(self)(self._name, Operators["eq"], self._value)], True
            ).pre_filter()
        return ""


class PatternFilter(FieldFilter):
    """Pattern filter, either a glob or a /regex/ pattern."""

    CLEAN_PRE_VAL_RE = re.compile(r"(?:\[.*?\])|(?:\(.*?\))|(?:{.*?})|(?:\\)")
    SPLIT_PRE_VAL_RE = re.compile(r"[^a-zA-Z0-9/_]+")
    SPLIT_PRE_GLOB_RE = re.compile(r"[?*]+")

    def validate(self):
        """Validate filter condition (template method)."""

        super().validate()
        self._value: str = self._value
        self._template = None
        self._flags = 0
        self._matcher: Callable[Any, Any]
        if (
            self._value == '""'
        ):  # Replace an empty string with a simple truthiness check
            self._matcher = lambda val, _: val == ""
        elif self._value.startswith("/") and (
            self._value.endswith("/") or self._value.endswith("/i")
        ):
            # Pick out a couple regexes that can be simplified
            if self._value in ["//", "/.*/", "//i", "/.*/i"]:
                self._matcher = lambda _, __: True
            elif self._value in ["/.+/", "/.+/i"]:
                self._matcher = lambda val, __: bool(val)
            else:
                value = self._value
                if self._value.endswith("/i"):
                    self._flags = re.IGNORECASE
                    value = self._value.rstrip("i")
                regex = re.compile(value[1:-1], self._flags)
                self._matcher = lambda val, _: regex.search(val)
        elif self._value.startswith("{{") or self._value.endswith("}}"):

            def _template_globber(val, item):
                """Helper method to allow templating a glob."""
                pattern = torrent.rtorrent.format_item(
                    torrent.rtorrent.env.from_string(self._template), item
                )
                return fnmatch.fnmatchcase(val, pattern)

            self._template = self._value
            self._matcher = _template_globber
        else:
            # Pick out a glob that can be simplified
            if self._value == "*":
                self._matcher = lambda _, __: True
            else:
                self._matcher = lambda val, _: fnmatch.fnmatchcase(val, self._value)

    def pre_filter_eq(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is None or self._template:
            return ""
        if not self._value or self._value == '""':
            return f'"equal={pf},cat="'

        if self._value.startswith("/") and (
            self._value.endswith("/") or self._value.endswith("/i")
        ):
            if self._value.endswith("/i"):
                needle = self._value[1:-2]
            else:
                needle = self._value[1:-1]
            needle = self.CLEAN_PRE_VAL_RE.sub(" ", needle)
            split_needle = self.SPLIT_PRE_VAL_RE.split(needle)
            # If the cleaning of the needle did not succeed and we still
            # have regex-y values, don't prefilter
            if any(metachar in needle for metachar in r"{}[]\\|()"):
                return ""
        else:
            split_needle = self.SPLIT_PRE_GLOB_RE.split(self._value)
        # Grab the longest needle available from the array
        needle = list(sorted(split_needle, key=len))[-1]

        if needle:
            # Skip trying to filter on non-ASCII characters
            try:
                needle.encode("ascii")
            except UnicodeEncodeError:
                return ""
            needle = needle.replace('"', r"\\\"")
            return r'"string.contains_i=${},\"{}\""'.format(pf, needle)

        return ""

    def eq(self, item):
        """Return True if filter matches item."""
        val = getattr(item, self._name) or ""
        result = self._matcher(val, item)
        return result


class FilesFilter(PatternFilter):
    """Pattern filter on filenames in a torrent."""

    def match(self, item) -> bool:
        """Return True if filter matches item. Overridden from the
        parent class to deal with with an array of strings rather than
        a single string
        """
        val = getattr(item, self._name)
        if val is not None:
            for fileinfo in val:
                if fnmatch.fnmatchcase(fileinfo.path, self._value):
                    return True
            return False
        return False


class TaggedAsFilter(FieldFilter):
    """Case-insensitive tags filter. Tag fields are white-space
    separated lists of tags.
    """

    def pre_filter_eq(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is not None:
            if self._exact and not self._value:
                return f'"equal={pf},cat="'
            if not self._value:
                return f'"not=${pf}"'
            return r'"string.contains_i=${},\"{}\""'.format(
                pf, self._value.replace('"', r"\\\"")
            )
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()
        self._value = self._value.lower()

        # If the tag starts with ':', test for exact equality (just this tag, no others)
        if self._value.startswith(":"):
            self._exact = True
            self._value = self._value[1:]
        else:
            self._exact = not self._value

    def eq(self, item) -> bool:
        """Return True if filter matches item."""
        tags = getattr(item, self._name) or []
        if self._exact:
            # Exact equality check
            return set(self._value) == set(tags)
        # Is given tag in list?
        return self._value in tags


class BoolFilter(FieldFilter):
    """Filter boolean values."""

    def pre_filter_eq(self):
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is not None:
            return '"equal={},value={}"'.format(pf, int(self._value))
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        self._value = truth(str(self._value), self._condition)
        self._condition = "yes" if self._value else "no"

    def eq(self, item):
        """Return True if filter matches item."""
        val = getattr(item, self._name) or False
        return bool(val) is self._value


class NumericFilterBase(FieldFilter):
    """Base class for numerical value filters."""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        self.not_null = False

    def match(self, item) -> bool:
        """Return True if filter matches item."""
        val = getattr(item, self._name) or 0
        if self.not_null and self._value and not val:
            return False
        # Grab the function from the native operator module
        op = getattr(operator, self._op.name)
        return bool(op(float(val), self._value))


def prefilter_field_lookup(name: str) -> Optional[str]:
    """Return the prefield field for a given name (if available)"""
    field = torrent.engine.field_lookup(name)
    if field is None:
        return None
    return field.prefilter_field


class FloatFilter(NumericFilterBase):
    """Filter float values."""

    FIELD_SCALE = dict(
        ratio=1000,
    )

    def pre_filter(self):
        """Prefilter a float value.

        rTorrent doesn't actually have floats, so we need to do a
        little translation for the prefiltering
        """
        pf = prefilter_field_lookup(self._name)
        if pf is None:
            return ""
        val = self._value * self.FIELD_SCALE.get(self._name, 1)
        lookup_table = {
            "eq": ("equal", int(val)),
            "ge": ("greater", math.floor(val - 1)),
            "gt": ("greater", math.floor(val)),
            "le": ("less", math.ceil(val + 1)),
            "lt": ("less", math.ceil(val)),
        }
        op, val = lookup_table[self._op.name]
        return f'"{op}=value=${pf},value={val}"'

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        try:
            self._value = float(self._value)
        except (ValueError, TypeError) as exc:
            raise FilterError(
                f"Bad numerical value {self._value!r} in {self._condition!r} ({exc})"
            ) from exc


class TimeFilter(NumericFilterBase):
    """Filter UNIX timestamp values."""

    TIMEDELTA_UNITS = dict(
        y=lambda t, d: t - d * 365 * 86400,
        M=lambda t, d: t - d * 30 * 86400,
        w=lambda t, d: t - d * 7 * 86400,
        d=lambda t, d: t - d * 86400,
        h=lambda t, d: t - d * 3600,
        m=lambda t, d: t - d * 60,
        s=lambda t, d: t - d,
    )
    TIMEDELTA_RE = re.compile(
        "^" + "".join(r"(?:(?P<{0}>\d+)[{0}{0}])?".format(i) for i in "yMwdhms") + "$"
    )

    # pylint: disable=super-init-not-called
    def __init__(self, name: str, op: FilterOperator, value: str):
        self.children = []
        self._name = name
        self.not_null = False
        self._condition = value
        self._op: FilterOperator = op
        self._duration = False

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is not None:
            # Adding a day of fuzz to avoid any possible timezone problems
            if self._op.name == "gt":
                timestamp = float(self._value) + 86400
                cmp_ = "greater"
            elif self._op.name == "lt":
                timestamp = float(self._value) - 86400
                cmp_ = "less"
            elif self._op.name == "eq":
                timestamp = 0
                cmp_ = "equal"
            else:
                return ""
            timestamp = float(self._value) + (
                -86400
                if self._op.name == "gt"
                else 86400
                if self._op.name == "lt"
                else 0
            )

            return '"{}=value=${},value={}"'.format(cmp_, pf, int(timestamp))
        return ""

    @property
    def _value(self):
        timestamp = now = time.time()

        if str(self._condition).isdigit():
            # Literal UNIX timestamp
            try:
                timestamp = float(self._condition)
            except (ValueError, TypeError) as exc:
                raise FilterError(
                    f"Bad timestamp value {self._condition!r} in {self._condition!r}"
                ) from exc
        else:
            # Something human readable
            delta = self.TIMEDELTA_RE.match(self._condition)
            if delta:
                # Time delta
                for unit, val in delta.groupdict().items():
                    if val:
                        timestamp = self.TIMEDELTA_UNITS[unit](timestamp, int(val, 10))

                if self._duration:
                    timestamp = now - timestamp
                # Invert the value for more intuitive matching (in
                # line with original code too) e.g. 'completed<1d'
                # should return things completed less than one day
                # *ago*
                else:
                    if self._op.name == "gt":
                        self._op = Operators["le"]
                    elif self._op.name == "ge":
                        self._op = Operators["lt"]
                    elif self._op.name == "lt":
                        self._op = Operators["ge"]
                    elif self._op.name == "le":
                        self._op = Operators["gt"]
            else:
                # Assume it's an absolute date
                if "/" in self._condition:
                    # U.S.
                    dtfmt = "%m/%d/%Y"
                elif "." in self._condition:
                    # European
                    dtfmt = "%d.%m.%Y"
                else:
                    # Fall back to ISO
                    dtfmt = "%Y-%m-%d"

                val = str(self._condition).upper().replace(" ", "T")
                if "T" in val:
                    # Time also given
                    dtfmt += "T%H:%M:%S"[: 3 + 3 * val.count(":")]

                try:
                    timestamp = time.mktime(tuple(time.strptime(val, dtfmt)))
                except (ValueError) as exc:
                    raise FilterError(
                        f"Bad timestamp value {self._condition!r} in {self._condition!r} ({exc})"
                    ) from exc

                if self._duration:
                    timestamp -= now
        return timestamp

    def validate(self):
        """Validate filter condition (template method)."""
        self._value  # pylint: disable=pointless-statement


class TimeFilterNotNull(TimeFilter):
    """Filter UNIX timestamp values, ignore unset values unless compared to 0."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_null = True
        self._duration = False


class DurationFilter(TimeFilter):
    """Filter durations in seconds."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_null = True
        self._duration = True

    def match(self, item) -> bool:
        """Return True if filter matches item."""
        if getattr(item, self._name) is None:
            # Never match "N/A" items, except when "-0" was specified
            return not bool(self._value)
        return super().match(item)


class ByteSizeFilter(NumericFilterBase):
    """Filter size and bandwidth values."""

    UNITS = dict(b=1, k=1024, m=1024**2, g=1024**3)

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        comparers = {
            "gt": "greater",
            "lt": "less",
            "eq": "equal",
        }
        pf = prefilter_field_lookup(self._name)
        if pf is not None and self._op.name in comparers:
            return '"{}={},value={}"'.format(
                comparers[self._op.name],
                pf,
                int(self._value),
            )
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        # Get scale
        lower_val = str(self._value).lower()
        if any(lower_val.endswith(i) for i in self.UNITS):
            scale = self.UNITS[lower_val[-1]]
            self._value = self._value[:-1]
        else:
            scale = 1

        # Get float value
        try:
            self._value = float(self._value)
        except (ValueError, TypeError) as exc:
            raise FilterError(
                f"Bad numerical value {self._value!r} in {self._condition!r}"
            ) from exc

        # Scale to bytes
        self._value = self._value * scale


QueryGrammar = Grammar(
    r"""
    query = (group / stmt) (ws (group / stmt))*
    stmt = (or_stmt / conds)
    or_stmt = (group / conds) ws or ws (group / conds)
    group = (not ws)? lpar ws stmt ws rpar
    conds = cond (ws cond)*
    cond = (&or / &lpar / &rpar / &not / named_cond / unnamed_cond)
    named_cond = word conditional filter
    unnamed_cond = filter
    filter      = (quoted_regex / quoted / glob / regex / word)
    glob = ~r"[\S]+"
    quoted_regex =  ~r'"/[^/\"]*/i?"'
    regex = ~r"/[^/]*/i?"
    quoted      = ~'"[^\"]*"'
    word        = ~r"[\w]+"
    conditional = (ne / ge / le / lt / gt / eq)
    ne          = ("!=" / "<>")
    eq          = ("==" / "=")
    gt          = (">" / "=+")
    lt          = ("<")
    ge          = (">=")
    le          = ("<=" / "=-")
    fws         = ~r"\s+"
    ws          = ~r"\s*"
    lpar = "["
    rpar = "]"
    or   = "OR"
    not  = ( "NOT" / "!" )
    """
)


class KeyNameVisitor(NodeVisitor):
    """Walk through a query tree and returns an array of key names.
    Implicit key names are not included."""

    # pylint: disable=unused-argument,missing-function-docstring
    def visit_expr(self, node, visited_children):
        output = ""
        for child in visited_children:
            output += str(child)
        return output

    def visit_named_cond(self, node, visited_children):
        return visited_children[0]

    def visit_word(self, node, visited_children):
        return [node.text]

    def visit_filter(self, node, visited_children):
        return ["name"]

    def generic_visit(self, node, visited_children):
        if visited_children:
            return [item for sublist in visited_children for item in sublist]
        return []


def create_filter(name: str, op: str, value: str):
    """Generates a filter class with the given name, operation and value"""
    field = torrent.engine.field_lookup(name)
    if field is None:
        raise SyntaxError(f"No such field '{name}'")
    filt = field._matcher
    return filt(name, op, value)


class MatcherBuilder(NodeVisitor):
    """Build a simplified tree of MatcherNodes to match an item against."""

    # pylint: disable=unused-argument,missing-function-docstring
    def visit_named_cond(self, node, visited_children):
        key, cond, needle = visited_children
        return create_filter(key, cond, needle)

    def visit_group(self, node, visited_children):
        return GroupNode(
            [c for c in visited_children[1:] if c is not None], visited_children[0]
        )

    def visit_not(self, node, visited_children):
        if node.text:
            return True
        return False

    def __pare_children(self, children, class_):
        """Get all non-None children, and if there's only one child left,
        return the child instead of wrapping it in the parent class.
        """
        real_children = [c for c in children if c is not None]
        if len(real_children) == 0:
            return None
        if len(real_children) == 1:
            return real_children[0]
        return class_(real_children)

    def visit_or_stmt(self, node, visited_children):
        return OrNode([c for c in visited_children if c is not None])

    def visit_conds(self, node, visited_children):
        if len(visited_children) == 2 and isinstance(visited_children[1], list):
            children = [visited_children[0]] + visited_children[1]
        else:
            children = visited_children
        pared = self.__pare_children(children, AndNode)
        return pared

    def visit_cond(self, node, visited_children):
        if len(visited_children) == 1 and isinstance(
            visited_children[0], (str, re.Pattern)
        ):
            return create_filter("name", Operators["eq"], visited_children[0])
        return self.generic_visit(node, visited_children)

    def visit_query(self, node, visited_children):
        return self.__pare_children(visited_children, AndNode)

    # Unfortunate but necessary boilerplate methods

    def visit_word(self, node, visited_children):
        return node.text

    def visit_quoted(self, node, visited_children):
        return node.text[1:-1]

    def visit_glob(self, node, visited_children):
        return node.text

    def visit_regex(self, node, visited_children):
        return node.text

    def visit_quoted_regex(self, node, visited_children):
        return node.text[1:-1]

    def visit_eq(self, node, visited_children):
        return Operators["eq"]

    def visit_ne(self, node, visited_children):
        return Operators["ne"]

    def visit_gt(self, node, visited_children):
        return Operators["gt"]

    def visit_ge(self, node, visited_children):
        return Operators["ge"]

    def visit_lt(self, node, visited_children):
        return Operators["lt"]

    def visit_le(self, node, visited_children):
        return Operators["le"]

    def generic_visit(self, node, visited_children):
        real_children = [c for c in visited_children if c is not None]
        if real_children:
            if isinstance(real_children, list) and len(real_children) == 1:
                return real_children[0]
            return real_children
        return None


RE_UNQUOTED_REGEX = re.compile(r"(\w*)(!?=)(\/.*\/)")


def cli_args_to_match_str(query: Sequence) -> str:
    """Convert CLI arguments to a string. Most usefully, this will
    automatically double quote unnamed conditions or regexes if they
    have a space in them.
    """
    args = []
    for a in query:
        if " " in a and not set("=><") & set(a) and not a.startswith('"'):
            a = f'"{a}"'
        if " " in a and RE_UNQUOTED_REGEX.match(a):
            a = RE_UNQUOTED_REGEX.sub(r'\1\2"\3"', a)
        args.append(a)
    return " ".join(args)


def create_matcher(query: Union[str, Sequence[str]]):
    """Utility function to build a matcher from a query string."""
    if not isinstance(query, str):
        query_str = cli_args_to_match_str(query)
    else:
        query_str = query
    return MatcherBuilder().visit(QueryGrammar.parse(query_str))



"""Move torrents to a new path.
Primarily intended for move-on-completion, but is intended to be generic.
"""
import shutil

from pathlib import Path

from pyrosimple.job import base
from pyrosimple.torrent import rtorrent


def move(i: rtorrent.RtorrentItem, target: str):
    """Move a torrent path"""
    i.stop()
    shutil.move(str(i.datapath()), target)
    i._engine.rpc.d.directory.set(i.hash, target)
    i.start()


class PathMover(base.MatchableJob):
    """Conditionally move torrent paths.
    Currently it just relies on the current datapath"""

    def __init__(self, config=None):
        """Set up statistics logger."""
        super().__init__(config)
        if not self.config["target"]:
            raise Exception("'target' not defined!")

    def run_item(self, item: rtorrent.RtorrentItem):
        """Conditionally move data"""
        target = rtorrent.format_item_str(self.config["target"], item)
        if not target:
            self.log.debug("Empty target for %s", item.hash)
            return
        if (Path(item.directory) == Path(target)) or (
            item.rpc_call("d.is_multi_file")
            and Path(item.directory).parent == Path(target)
        ):
            self.log.debug("%s already moved, skipping", item.hash)
            return
        if self.config["dry_run"]:
            self.log.info(
                "Would move %s from '%s' to '%s'", item.hash, item.datapath(), target
            )
            return
        self.log.info(
            "Moving path for %s from '%s' to '%s'", item.hash, item.datapath(), target
        )
        move(item, target)
        item.flush()
